;;  ReverseDirection.lsp [command name: RD]
;;  To Reverse the Direction of objects with directionality, usually for the
;;    purpose of having embedded text or shapes in complex linetypes "face"
;;    the right way when that is dependent on the objects' drawn direction.
;;  Works on Lines, Arcs, Circles, Polylines [any variety, including curve-fitted
;;    2D and splined 2D or 3D], Ellipses, Splines & Mlines.
;;  Preserves all properties of selected objects [Layer, Space, Layout, any over-
;;    rides of Color, Linetype, Linetype Scale, Lineweight, Thickness].
;;  Converts Arcs, Circles, & Ellipses with minor-major axis ratio of 1 [circular]
;;    to LWPolylines without asking for User consent, since making different
;;    entity type is only way to reverse direction, and geometry is unaffected.
;;  Converts clockwise Polylines that are donut-like circular or of single arc
;;    segment, and global zero width, to Circles/Arcs [see instructions within
;;    (rd-lwp) to retain as Polylines].
;;  Asks User consent to convert Ellipses [with axis ratio < 1] to Splines, and
;;    to generate new fit points for certain Splines; geometry is altered, to User-
;;    specified precision [number of segments].  Retains precision [separately
;;    for Ellipses and Splines] and offers as default on subsequent use.
;;  For Ellipses [with axis ratio < 1], Spline approximation is far more accurate
;;    with fit points AT Ellipse's quadrant points and variably spaced [i.e. closer
;;    together as curvature gets tighter]; to achieve this, User-specified number of
;;    segments will sometimes be increased slightly.  [Spline approximation is
;;    also more accurate than Polyline with PELLIPSE = 1 at 16 segments that a
;;    Polyline Ellipse would use, and far more accurate with higher precision.]
;;  For 2D Polylines [lightweight or "heavy"], retains global or varying width(s).
;;  For 3D Polylines, reverses even though they do not honor non-continuous
;;    linetypes, in case they may in future, or in case of other reasons to want
;;    them reversed.
;;  For Multilines, if they have been MLEdited in relation to other Mlines, that
;;    will need to be redone; RD only draws overall Mline reversed.
;;  User option to Undo latest reversal, individually within command, so that
;;    e.g. Ellipse/Spline can be re-reversed with higher precision if desired.
;;  Note:  in Polylines of complex linetypes with VARYING width, embedded
;;    text or shapes, in ARC segments only, seem to sometimes, but not always,
;;    be fixed facing counterclockwise, and to NOT be reversible.  Those in
;;    LINE segments are always reversed [thus facing opposite any un-reversed
;;    in arc segments], as are those in BOTH line and arc segments of GLOBAL-
;;    width Polylines, WITH linetype generation enabled.  Reversal or not in arc
;;    segments seems to be related to particular geometry and/or linetype scale;
;;    can rarely reverse in some, but not all, arc segments in the same Polyline.
;;  [Written in AutoCAD 2004 -- some circumstances above may have changed
;;    in later versions.]
;;  Kent Cooper, last edited 24 January 2014

(defun C:RD ; = Reverse Direction
  (/ *error* rd-lwp getprec cmde n esel ent edata etype clo dxfpts pt verts lastwb
    len int vert vdata parA parZ precq parinc parcum seg mlvals mlnames done)
    ; Some variable names used for different purposes / in different
    ;   ways in reversing different entity types [pt len int dxfpts].
    ; Global variables retained for subsequent use:
    ;   *eprec *sprec [in Ellipse and no-fit-point Spline reversals].

  (defun *error* (errmsg)
    (if (not (wcmatch errmsg "Function cancelled,quit / exit abort,console break"))
      (princ (strcat "\nError: " errmsg))
    ); if
    (mapcar 'setvar svnames svvals)
    (if mlvals (mapcar 'setvar mlnames mlvals)); in case of failure within Mline reversal
    (command "_.undo" "_end")
    (setvar 'cmdecho cmde)
    (princ)
  ); defun - *error*

  (defun rd-lwp (/ verts lastwb)
    ; = Reverse Direction of LWPolyline -- for LWPline, & 2DPline/Arc after conversion to LWP
    (cond
      ; If two-semicircle-arc-segment circular LWP, 0 width, clockwise curvature, converts to
      ; Circle to reverse direction.  To keep as LWP, remove following condition entirely.
      (; circular [e.g. from RD reversal of a Circle] -- convert [back] to Circle
        (and
          clo
          (member '(90 . 2) edata); 2 vertices
          (member '(43 . 0.0) edata); global zero width
          (equal (cdr (assoc 42 edata)) -1.0 1e-8); both segments CW semi-circles
          (equal (cdr (assoc 42 (reverse edata))) -1.0 1e-8)
            ; done via (equal) with fuzz factor, instead of (= [both 42 entries] -1.0), because the
            ; latter can miss one converted from circular/arc Ellipse
        ); and
        (setq pt ; center of LWPline circle
          (mapcar '/
            (mapcar '+ (vlax-curve-getStartPoint ent) (vlax-curve-getPointAtParam ent 1))
            '(2 2 2)
          ); mapcar /
        ); setq
        (entmake
          (append ; circle-specific entries
            '((0 . "CIRCLE") (100 . "AcDbEntity") (100 . "AcDbCircle"))
            (vl-remove-if-not '(lambda (x) (member (car x) '(67 410 8 62 6 48 370 39))) edata)
              ; remove Polyline-specific entries from Polyline's entity data,
              ; and extrusion direction; keep space/tab/layer, any overrides
            (list
              (cons 10 (trans pt 0 ent)); center -- WCS to OCS
              (cons 40 (distance pt (vlax-curve-getStartPoint ent))); radius
              (assoc 210 edata); extr. dir. at end [if in middle, reverts to (210 0.0 0.0 1.0) in (entmake)]
            ); list
          ); append
        ); entmake
        (entdel ent)
      ); circular condition
      ; If one-arc-segment LWP, 0 width, clockwise curvature, converts to Arc to
      ; reverse direction.  To keep as LWP, remove following condition entirely.
      (; one-arc-segment CW
        (and
          (not clo); open
          (member '(90 . 2) edata); 2 vertices [1 segment]
          (member '(43 . 0.0) edata); global zero width
          (minusp (cdr (assoc 42 edata))); clockwise arc
        ); and
        (command "_.explode" ent ""); then -- convert to Arc [will be CCW]
      ); arc condition
      (T ; not CW circular or single-arc-segment -- reverse it
        (setq
          verts
          (reverse ; so below transformation works in already-reversed vertex order
            (vl-remove-if-not
              '(lambda (x)
                (member (car x) '(10 40 41 42)); keep vertex locations, widths, bulges
              );  lambda
              edata
            ); vl-remove-if-not
          ); reverse & verts
          lastwb ; new last-segment widths/bulge to move to end [values matter only if closed]
          (list
            (cons 40 (cdadr verts)); starting width
            (cons 41 (cdaddr verts)); ending width
            (cons 42 (- (cdar verts))); opposite-side bulge factor
          ); list
          verts (cdddr verts); to start next with new first vertex
        ); setq
        (repeat (/ (length verts) 4)
          (setq dxfpts
            (append
              dxfpts
              (list
                (car verts); vertex location
                (cons 40 (cdaddr verts)); new starting width
                (cons 41 (cdr (assoc 40 verts))); new ending width [(cdadddr) not defined]
                (cons 42 (- (cdadr verts))); opposite-side bulge factor
              ); list
            ); append
            verts (cddddr verts); on to next set
          ); setq
        ); repeat
        (entmod
          (append
            (vl-remove-if '(lambda (x) (member (car x) '(10 40 41 42 210))) edata)
              ; removes vertex elements, extrusion direction; keeps entity type/
              ; names, handle, global width if any, space, layer, layout, no. of
              ; vertices, open/closed, linetype generation, elevation, any overrides
            dxfpts
            verts ; what's left of it [only original first vertex = new last one]
            lastwb; past-last-vertex widths/bulge
            (list (assoc 210 edata)) ; extr. dir. at end [if in middle, reverts to (210 0.0 0.0 1.0) in (entmake)]
          ); append
        ); entmod
      ); other LWPolyline condition
    ); cond
  ); defun -- rd-lwp

  (defun getprec (/ prec); = GET PRECision [# of segments]; for Ellipse/Spline
    (setq prec (read (strcat "*" (strcase (substr etype 1 1) T) "prec")))
      ; = Ellipse or Spline reversal PRECision [*eprec or *sprec]
    (set prec
      (cond
        ((initget 6)); no 0, no negative; returns nil
        ( (getint ; [returns nil on Enter]
            (strcat
              "\nPrecision [number of segments] <"
              (if (eval prec) (itoa (eval prec)) "24"); offer 24 default on first use
              ">: "
            ); strcat
          ); getint
        ); User entry condition
        ((eval prec)); prior value, if any
        (24); Enter on first use
      ); cond & *eprec/*sprec
    ); set
  ); defun - getprec

  (vl-load-com)
  (setq cmde (getvar 'cmdecho))
  (setvar 'cmdecho 0)
  (command "_.undo" "_begin")
  (setq
    n 0 ; for Undo option
    svnames ; = System Variable NAMES
      '(osmode blipmode cecolor celtype celtscale
      celweight thickness peditaccept ucsfollow)
    svvals (mapcar 'getvar svnames); = System Variable VALueS
  ); end setq
  (mapcar 'setvar svnames '(0 0 "BYLAYER" "BYLAYER" 1.0 -1 0.0 1 0))
    ; need to set first 7 above to defaults, or if any current settings are different,
    ; things made with (entmake) will carry those, and if not overridden
    ; in selected object, won't be fixed because not present in entity data
  (setvar 'plinetype 2)
    ; assuming no reason to save and set back if <2, but could do that
    ; needs to be 1 or 2 so Arc conversion will be LWPolyline

  (while (not done)
    (setvar 'errno 0)
    (if (> n 0) (initget "Undo")); Undo option only if things to undo
    (setq esel
      (entsel
        (strcat
          "\nSelect object to reverse, or "
          (if (> n 0) "[Undo] " ""); Undo option only if things to undo
          "<exit>: "
        ); strcat
      ); entsel
    ); setq
    (cond ; overall object-selection or Undo option
      ((= esel "Undo")
        (setq n (1- n))
        (command "_undo" "_back"); revert last selected object to original direction
      ); Undo condition
      ( ; test for selection, valid entity type, unlocked Layer
        (and
          (= (getvar 'errno) 0); picked something
          (setq
            ent (car esel)
            edata (entget ent)
          ); setq
          (wcmatch (setq etype (cdr (assoc 0 edata))) "*LINE,ARC,CIRCLE,ELLIPSE")
            ; may not be comprehensive list of valid entity types for later versions
          (/= etype "XLINE"); disallowed type that *LINE above lets in
          (= (logand 4 (cdr (assoc 70 (tblsearch "layer" (cdr (assoc 8 edata)))))) 0)
            ; 0 = unlocked, 4 = locked
        ); and
        (command "_.undo" "_mark")
        (setq
          clo ; = CLOsed; used only in (rd-lwp)/Ellipse/Spline/Mline
            (if (= etype "MLINE")
              (= (logand (cdr (assoc 71 edata)) 2) 2); then [Mlines aren't "curves"]
              (vlax-curve-isClosed ent); else
            ); if
          dxfpts nil ; used only in (rd-lwp)/3DPline/Ellipse/Spline/Mline
            ; reset for subsequent objects in same RD command, where applicable
        ); setq
        (setvar 'blipmode 0)

        (cond ; VALID SELECTION -- entity-type options

          ((= etype "LINE")
            (setq
              pt (cdr (assoc 10 edata)); starting endpoint
              edata (subst (cons 10 (cdr (assoc 11 edata))) (assoc 10 edata) edata)
              edata (subst (cons 11 pt) (assoc 11 edata) edata)
            ); setq
            (entmod edata)
          ); Line condition

          ((= etype "ARC")
            (command
              "_.ucs" "_new" "_object" ent ; won't Pedit if not parallel to current CS
              "_.pedit" ent "" ; convert to LWPolyline
              "_.ucs" "_previous"
            ); command
            (setq ent (entlast) edata (entget ent)); re-establish for (rd-lwp) to use
            (rd-lwp)
          ); Arc condition

          ((= etype "CIRCLE")
            (setq
              pt (cdr (assoc 10 edata)); center point, OCS for Circle & LWPolyline w/ WCS 0,0,0 as origin
              len (cdr (assoc 40 edata)); radius
            ); setq
            (entmake
              (append ; add Polyline-specific entries
                '((0 . "LWPOLYLINE") (100 . "AcDbEntity") (100 . "AcDbPolyline"))
                (vl-remove-if '(lambda (x) (member (car x) '(-1 0 330 5 100 10 40 210))) edata)
                  ; remove Circle-specific entries and extrusion direction from
                  ; entity data; keep entries present only if not default/bylayer
                (list
                  '(90 . 2); # of vertices
                  (cons 70 (1+ (* 128 (getvar 'plinegen))))
                    ; closed [the 1], and uses current linetype-generation setting; change above
                    ; line to '(70 . 129) to force linetype generation on, '(70 . 1) to force it off
                  '(43 . 0.0); global zero width
                  (cons 38 (caddr pt)); elevation in OCS above WCS origin [Z of Circle center]
                  (list 10 (+ (car pt) len) (cadr pt)); vertex 1
                  '(40 . 0.0) '(41 . 0.0) '(42 . -1.0); 0 start & end widths, clockwise semi-circle bulge factor
                  (list 10 (- (car pt) len) (cadr pt)); vertex 2
                  '(40 . 0.0) '(41 . 0.0) '(42 . -1.0)
                  (assoc 210 edata) ; extr. dir. at end [if in middle, reverts to (210 0.0 0.0 1.0) in (entmake)]
                ); list
              ); append
            ); entmake
            (entdel ent)
          ); Circle condition

          ((= etype "LWPOLYLINE")
            (rd-lwp)
          ); LWPolyline condition

          ((= etype "POLYLINE"); "heavy" 2D or 3D
            (setq
              etype (substr (cdr (assoc 100 (reverse edata))) 5 2); replace with "2d" or "3d"
              int (vla-get-Type (vlax-ename->vla-object ent)); type of 2D or 3D curvature
            ); setq
            (if (> int 0)
              ; curve-fit [2D only] or cubic-or-quadratic splined [either kind];
              ; int = 0 if not curve-fit/splined, including 2D with arc segments
              (progn
                (if (= etype "2d") (command "_.ucs" "_new" "_object" ent))
                  ; won't Pedit 2D if not parallel to current CS, but can't align UCS with 3D
                (command "_.pedit" ent "_decurve" ""); 2D or 3D; retains entity name, handle;
                  ; must do it this way w/ 2D to make LWPoly [putting Type 0 leaves it 2D];
                  ; can't do it to all 2D: even w/o curve-fit/spline, arc segments straighten
                (if (= etype "2d") (command "_.ucs" "_previous"))
              ); progn
            ); if
            (if (= etype "2d")
              (progn ; then -- 2D
                (command "_.convertpoly" "_light" ent ""); retains entity name, handle
                (setq edata (entget ent)); replace with LW-version data for (rd-lwp) to use
                (rd-lwp)
                ; Following returns to 2DPline as when selected, even if not curve-fit/splined,
                ; and if not reverted to Circle/Arc; to leave LWPline, eliminate following (if).
                (if (= (cdr (assoc 0 (entget ent))) "LWPOLYLINE")
                  (command "_.convertpoly" "_heavy" ent "")
                ); if
              ); progn -- 2D
              (progn ; else -- 3D
                (setq vert ent) ; must retain ent for (entdel) below
                (while
                  (= (cdr (assoc 0 (setq vdata (entget (setq vert (entnext vert)))))) "VERTEX")
                  (setq dxfpts (cons (trans (cdr (assoc 10 vdata)) 0 1) dxfpts)); cons builds in reverse
                ); while
                (if (vlax-curve-isClosed ent) (setq dxfpts (append dxfpts (list "_close"))))
                (command "_.3dPoly" (apply 'command dxfpts))
                (entmod 
                  (append
                    (entget (entlast))
                    (vl-remove-if-not '(lambda (x) (member (car x) '(8 62 6 48 370))) edata)
                      ; impose original's Layer, any overrides [supersede current settings]
                  ); append
                ); entmod
                (entdel ent)
              ); progn
            ); if
            (if (> int 0); original was curve-fit [2D only] or splined [either]
              (vla-put-Type (vlax-ename->vla-object (if (= etype "2d") ent (entlast))) int)
                ; re-impose specific type of curve
            ); if
          ); 2D/3D Polyline condition

          ((= etype "ELLIPSE")
            (if (equal (cdr (assoc 40 edata)) 1.0 1e-8)
              ; e.g. from Circle/Arc in exploded Block with X=Y but different Z]
              ; [simpler (member '(40 . 1.0) edata) can miss via tiny calculation variance]
              (progn ; then
                (command
                  "_.ucs" "_new" "_object" ent
                  "_.pline"
                    (trans (vlax-curve-getEndPoint ent) 0 1)
                    "_width" 0 0 "_arc" "_second"
                    (trans
                      (vlax-curve-getPointAtDist ent
                        (* (vlax-curve-getDistAtParam ent (vlax-curve-getEndParam ent)) 0.75)
                          ; quadrant in clockwise direction for full-circle Ellipse
                      ); getPoint
                      0 1
                    ); trans
                    (trans
                      (if clo
                        (vlax-curve-getPointAtParam ent pi); then - circle
                        (vlax-curve-getStartPoint ent); else -- arc
                      ); if
                      0 1
                    ); trans
                    (if clo "_close" ""); circle vs. arc -- end Pline command
                  "_.ucs" "_previous"
                ); command
                (entmod
                  (append
                    (entget (entlast))
                    (vl-remove-if-not '(lambda (x) (member (car x) '(8 62 6 48 370))) edata)
                      ; impose original's Layer, any overrides [supersede current settings]
                  ); append
                ); entmod
                (entdel ent)
              ); progn
              (progn ; else -- not circular [minor-major axis ratio < 1]
                (initget "Yes No")
                (if
                  (=
                    (getkword "\nConvert to Spline approximation to reverse? [Y/N] <N>: ")
                    "Yes"
                  ); =
                  (progn ; then
                    (getprec)
                    (setq
                      parA (vlax-curve-getStartParam ent)
                      parZ (vlax-curve-getEndParam ent)
                      precq (/ (* pi *eprec) (- parZ parA) 2)
                        ; = PRECision density per Quadrant; then if not whole, round up:
                      precq (+ (fix precq) (if (equal (rem precq 1.0) 0.0 1e-8) 0 1))
                      parinc (/ pi 2 precq); = PARameter INCrement to step around Ellipse
                    ); setq
                    (if (equal (rem parA parinc) 0.0 1e-8); starting at a parameter-increment point
                      (setq parcum (- parA parinc)); then
                        ; PARameter CUMulative value along Ellipse backed off one increment for starting basis
                      (setq ; else
                        parcum (* parinc (fix (/ parA parinc))); rounded down for starting basis
                        dxfpts (list (cons 11 (vlax-curve-getStartPoint ent))); start point as first Spline fit point
                      ); setq
                    ); if
                    (while (< (+ parcum parinc) parZ)
                      (setq dxfpts
                        (cons
                          (cons 11 (vlax-curve-getPointAtParam ent (setq parcum (+ parcum parinc))))
                          dxfpts
                        ); cons
                      ); setq
                    ); while
                    (if (not (equal parcum parZ 1e-8)); cumulative total didn't reach end
                      (setq dxfpts ; add last fit point
                        (cons
                          (cons 11 (vlax-curve-getEndPoint ent))
                          dxfpts
                        ); cons
                      ); setq
                    ); if
                    (entmake
                      (append
                        '((0 . "SPLINE") (100 . "AcDbEntity") (100 . "AcDbSpline"))
                        (vl-remove-if '(lambda (x) (member (car x) '(-1 0 330 5 100 10 11 40 41 42))) edata)
                          ; removes entity name/types, handle, center, eccentricity, start/end
                          ; angles; keeps space, layer, layout, extrusion direction, any overrides
                        (list
                          (cons 70 (if clo 11 8))
                          '(71 . 3); degree -- apparently default value
                          '(42 . 1.0e-010) '(43 . 1.0e-010) '(44 . 1.0e-010); tolerances -- default values
                          (cons 12 (mapcar '- (vlax-curve-getFirstDeriv ent (vlax-curve-getEndParam ent)))); start tangent
                          (cons 13 (mapcar '- (vlax-curve-getFirstDeriv ent (vlax-curve-getStartParam ent)))); end tangent
                        ); list
                        dxfpts ; fit points
                        ; DXF codes 72, 73, 74, 40's & 10's expendable -- resultant from other values
                      ); append
                    ); entmake
                    (entdel ent)
                  ); progn -- then
                ); if -- permission to convert to Spline or not
              ); progn -- not circular
            ); if -- minor-major ratio = 1.0 or not
          ); Ellipse condition

          ((= etype "SPLINE")
            (if (assoc 11 edata)
              (progn ; then -- fit points exist [as drawn using Spline command, or from RD reversal]
                (entmod
                  (append
                    (vl-remove-if '(lambda (x) (member (car x) '(12 13 40 10 11))) edata)
                      ; 10 & 40 entries are dispensable if there are 11 entries [fit points]
                    (if (assoc 13 edata) (list (cons 12 (mapcar '- (cdr (assoc 13 edata)))))); new start tangent from old end
                    (if (assoc 12 edata) (list (cons 13 (mapcar '- (cdr (assoc 12 edata)))))); new end tangent from old start
                    (reverse (vl-remove-if-not '(lambda (x) (= (car x) 11)) edata)); reverse sequence of fit points
                  ); append
                ); entmod
              ); progn -- reverse fit points
              (progn ; else -- no fit points [via Offset (incl. of Ellipse)/Trim/Break]
                (initget "Yes No")
                (if
                  (=
                    (getkword "\nMust approximate to reverse -- proceed? [Y/N] <N>: ")
                    "Yes"
                  ); =
                  (progn
                    (getprec)
                    (setq
                      len (vlax-curve-getDistAtParam ent (vlax-curve-getEndParam ent)); = LENgth
                      seg (/ len *sprec); = SEGment length for fit-point spacing
                      int -1
                    ); setq
                    (repeat (1+ *sprec) ; list of (assoc 11) entries for fit points
                      (setq dxfpts
                        (cons ; builds list already in reversed order
                          (cons 11 (vlax-curve-getPointAtDist ent (* seg (setq int (1+ int)))))
                          dxfpts
                        ); cons
                      ); setq
                    ); repeat
                    (entmod
                      (append
                        (vl-remove-if '(lambda (x) (member (car x) '(72 73 74 12 13 10 11 40))) edata)
                          ; removes numbers of points, tangents, control & fit points, knots; keeps entity type/
                          ; names, handle, flag, layer, layout, extrusion direction, degree, tolerances, any overrides
                        (list (cons 74 (+ *sprec (if clo 1 0)))); # of fit points
                        (if clo ; open Splines w/o fit points don't have 12/13
                          (list
                            (cons 12 (mapcar '- (vlax-curve-getFirstDeriv ent (vlax-curve-getStartParam ent))))
                            (cons 13 (mapcar '- (vlax-curve-getFirstDeriv ent (vlax-curve-getEndParam ent))))
                          ); list
                        ); if
                        dxfpts ; fit points [11] from (rd-spl11)
                      ); append
                    ); entmake
                  ); progn -- proceed to reverse
                ); if -- Yes/No
              ); progn -- else
            ); if -- has fit points or not
          ); Spline condition

          ((= etype "MLINE")
            ; Multilines honor non-continuous linetypes assigned when Bylayer,
            ; and within defined Multiline Styles, but not linetype overrides
            (setq mlvals
              (mapcar 'getvar (setq mlnames '(cmljust cmlscale cmlstyle)))
            ); setq
            (command
              "_.ucs" "_new" "_zaxis"
                ; align with selected Mline [UCS New OBject doesn't always work as expected]
              (trans (cdr (assoc 11 edata)) 0 1); base = first definition point
              (trans (mapcar '+ (cdr (assoc 11 edata)) (cdr (assoc 210 edata))) 0 1); + extr dir from there
            ); command
            (setq dxfpts
              (mapcar '(lambda (x) (trans x 0 1))
                (mapcar 'cdr ; strip 11's out of definition point entries
                  (reverse (vl-remove-if-not '(lambda (x) (= (car x) 11)) edata))
                ); mapcar 'cdr
              ); mapcar [trans] & dxfpts
            ); setq
            (if clo (setq dxfpts (append dxfpts (list "_close"))))
            (setvar 'cmljust (- 2 (cdr (assoc 70 edata))))
              ; toggles between Top [0] & Bottom [2], leaves Zero [1] unchanged
            (setvar 'cmlscale (cdr (assoc 40 edata)))
            (setvar 'cmlstyle (cdr (assoc 2 edata)))
            (command
              "_.mline" (apply 'command dxfpts)
              "_.ucs" "_previous"
            ); command
            (entmod
              (append
                (entget (entlast))
                (vl-remove-if-not '(lambda (x) (member (car x) '(8 62 6 48 370))) edata)
                  ; impose original's Layer, any overrides [supersede current settings]
                  ; including linetype if present, even though not honored in display
              ); append
            ); entmod
            (mapcar 'setvar mlnames mlvals); reset Mline-specific system variables
            (setq mlvals nil); so *error* doesn't need to reset on failed reversal of other type
            (entdel ent)
          ); Mline condition

        ); cond -- entity types

        (setq n (1+ n)); for Undo option

      ); picked-something-valid condition

      ((= (getvar 'errno) 0); picked something, but invalid under (and) tests above
        (prompt "\nCannot be reversed [inapplicable type or on locked Layer] --")
      ); non-qualifying condition

      ((= (getvar 'errno) 7) (prompt "\nNothing selected --"))

      ((setq done T)); Enter/space at Select-object prompt [errno = 52]; stop (while) loop

    ); cond -- selection or U
  ); while
  (mapcar 'setvar svnames svvals); reset shared default/bylayer system variables
  (command "_.undo" "_end")
  (setvar 'cmdecho cmde)
  (princ)
); defun -- RD
(prompt "\nType RD to Reverse the Direction of directional objects.")